package graphomat.graph;
import java.io.Serializable;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Observer;
import java.util.Observable;
import graphomat.util.*;
/**
 *This class provides the possiblity to store nodes and their connections.
 *The edges and the nodelist can be traversed with the
 *<code>Enumeration</code> given by <code>getNodeEnum()</code> and
 *<code>getEdgeEnum()</code>.There is a second posibility
 *with the <code>TraversNodes()</code> and the equivalent
 *<code>TraversEdges()</code> method.
 */

public class Graph extends ObservSerializable implements Observer{
	static final long serialVersionUID = -602562576492697539L;
	private final static int HASH_INITIALCAPACITY = 10;
	private DLinkedList nodeList = new DLinkedList();
	private DLinkedList edgeList = new DLinkedList();
	private SubGraph subG = new SubGraph(this);
	public static final Object NODEADDED = new Object();
	public static final Object NODEMOVED = new Object();
	public static final Object NODEDELETED = new Object();
	public static final Object NODESTATECHANGE = new Object();
	public static final Object EDGEADDED = new Object();
	public static final Object EDGEDELETED = new Object();
	
	public Graph() {
		nodeList.addObserver(this);
		edgeList.addObserver(this);
	}
	
 /**
  *All nodes added to the Graph have to be initallized first with the
  *<code>new</code> operator, as shown below:
  *<Blockquote><code>
  *Graph g = new Graph();<br>
  *g.addNode(new Node(4,1));
  *</code></Blockquote>
  *Uninitalized nodes will be ignored.
  */
	public void addNode(Node n) {
		if (n != null) {
			nodeList.insertElement(n);
			n.setConnector(new Connector(n, edgeList, HASH_INITIALCAPACITY));
			subG.addNode(n);
			notifyObservers(NODEADDED);
			n.addObserver(this);
		} else System.out.println("Warning : tried to add an unintialized Node."
		                          +" Method-call ignored!");
	}

 /**
  *Nodes deleted from the Graph will be disconnected first.
  *Uninitalized nodes will be ignored.
  *<Blockquote><code>
  *Graph g = new Graph();<br>
  *Node n1 = new Node();<br><br>
  *g.addNode(n1);<br>
  *g.deleteNode(n1);<br>
  *</code></Blockquote>
  *Uninitalized nodes will be ignored.
  */
	public void deleteNode(Node n) {
	if (n != null) {
			n.disconnectAll();
			n.setSelected(false);
			nodeList.deleteElement(n);
			n.notifyObservers(Node.DELETED);
		} else System.out.println("Warning : tried to delete an unintialized Node."
		                        + " Method-call ignored!");
	}
	
	/**
	 *Checks if the graphs owns the requested node.
	 */	
	public boolean inGraph(Node n) {return nodeList.inList(n);}

	/**
	 *Returns the current Selection of the Graph. If nothing is selected the
	 *whole Graph will be returned as a <code>SubGraph</code>.
	 */
	public SubGraph getSelection() {return subG;}
	
	/**
	 *Returns the node with a given location. Is usefull if u forgot the
	 *handle of a specific node. If there are several nodes at this location,
	 *the oldest will be returned. <b>Warning:</b> <code>getFirstNodeAt()</code>
	 *returns null if there was no Node at this location.
   *<Blockquote><code>
   *Graph g = new Graph();<br>
   *Node n;<br><br>
   *g.addNode(new Node(4,1));<br>
   *n = g.getFirstNodeAt(4,1));
   *</code></Blockquote>
	 */ 
	public Node getFirstNodeAt(double x, double y) {
		Enumeration enum = nodeList.getEnum();
		Node n;

		while (enum.hasMoreElements()) {
			n = (Node) enum.nextElement();
			if (n.samePlace(x,y))
				return n;
		}
		return null;
	}

	/**
	 *Returns the Edges connecting the two Nodes.
	 */
	public Edge getEdge(Node n1, Node n2)	{return n1.getConnector().getEdge(n2);}

	/**
	 *Traverses all nodes and processes it with the given
	 *<code>NodeTraverser</code>.
	 *<Blockquote><code>
	 *class MoveTraverser extends NodeTraverser {<br>
	 *<spacer size=25>public void processNode(Node n) {<br>
	 *<spacer size=50>n.move(-1,2);<br>
	 *<spacer size=25>}<br>
	 *}<br><br>
	 *Graph g = new Graph();<br>
	 *MoveTraverser trav = new MoveTraverser();<br><br>
	 *g.addNode(new Node(2,3));<br>
	 *g.addNode(new Node(1,0));<br>
	 *g.traversNodes(trav);
   *</code></Blockquote>
	 *There is also the possibility to define a <code>NodeTraverser</code>
	 *on-the-fly with a anonymous class.
	 *<Blockquote><code>
	 *Graph g = new Graph();<br>
	 *g.addNode(new Node(2,3));<br>
	 *g.addNode(new Node(1,0));<br>
	 *g.traversNodes(new NodeTraverser() {<br>
	 *<spacer size=25>public void processNode(Node n) {<br>
   *<spacer size=50>n.move(-1,2);<br>
   *<spacer size=25>}<br>
   *});<br>
   *</code></Blockquote>
	 */
	public void traversNodes(NodeTraverser t) {
		Enumeration enum = getNodeEnum();
		while(enum.hasMoreElements()) {
			t.processNode((Node)enum.nextElement());
		}
	}

	/**
 	 *Traverses all edges and processes it with the given
	 *<code>EdgeTraverser</code>.
	 *@see traversNodes
	 */
	public void traversEdges(EdgeTraverser t) {
		Enumeration enum = getEdgeEnum();
		while(enum.hasMoreElements()) {
			t.processEdge((Edge)enum.nextElement());
		}
	}

	/**
	 *Returns a <code>Enumeration</code> containing the nodes of the Graph.
	 *<Blockquote><code>
	 *Graph g = new Graph();<br>
	 *Enumertaion enum;<br><br>
	 *g.addNode(new Node(2,3));<br>
	 *g.addNode(new Node(1,0));<br>
	 *enum = g.getEnum();<br>
	 *while(enum.hasMoreElements()) {<br>
	 *<spacer size=25>((Node)enum.nextElement()).move(-1,2);<br>
	 *}<br>
   *</code></Blockquote>
	 */
	public Enumeration getNodeEnum() {return nodeList.getEnum();}

	/**
	 *Returns a <code>Enumeration</code> containing the edges of the Graph.
	 *@see getNodeEnum
	 */
	public Enumeration getEdgeEnum() {return edgeList.getEnum();}

	/**
	 *Returns the number of nodes in the Graph.
	 */
	public int getNodeCount()	{return nodeList.getElementCount();}

	/**
	 *Returns the number of edges in the Graph.
	 */
	public int getEdgeCount()	{return edgeList.getElementCount();}

	/**
	 *Returns a String discribing the whole Graph.
	 */
	public String toString() {return nodeList.toString();}
	
	public void notifyObservers(Object arg) {
		setChanged();
		super.notifyObservers(arg);
	}
	
	public void update(Observable n, Object arg) {		
		if (n == nodeList) {
			if (arg == DLinkedList.DELETED) notifyObservers(NODEDELETED);
		} else if (n == edgeList) {
			if (arg == DLinkedList.ADDED) notifyObservers(EDGEADDED);
			if (arg == DLinkedList.DELETED) notifyObservers(EDGEDELETED);
		} else if (n instanceof Node) {
			if (arg == Node.MOVED) notifyObservers(NODEMOVED);
			if (arg == Node.STATECHANGE) notifyObservers(NODESTATECHANGE);
		}
	}
}
