package graphomat.graph;
import graphomat.util.*;
import java.io.*;
import java.util.Enumeration;
import java.util.Observer;
import java.util.Observable;

public class SubGraph extends ObservSerializable implements Observer{
	abstract class Getter implements Serializable{
		abstract public double Get(Node n);
	}
	class Getx extends Getter {
		public double Get(Node n) {return n.getx();}
	}
	class Gety extends Getter {
		public double Get(Node n) {return n.gety();}
	}	
	class MinValue extends LazyDouble{
		private Node n;
		private HashSet h = new HashSet();
		private Getter g;
		
		public MinValue(HashSet h, Getter g)	{
			this.h = h;
			this.g = g;
		}
		public void setNode(Node n) {
			this.n = n;
			setValue(g.Get(n));
		}
		public Node getNode()	{return n;}
		public boolean Compare(double x1, double x2) {return (x1 < x2);}
		public void recalculate() {
			Node n;
			double x;
			if (h.getCount() > 0) {
				Enumeration enum = h.getEnum();
				n = (Node) enum.nextElement();
				x = g.Get(n);
				setNode(n);
				while (enum.hasMoreElements()) {
					n = (Node) enum.nextElement();
					if (Compare(g.Get(n),x)) {x = g.Get(n); setNode(n);}
				}
			}
		}
	}
	class MaxValue extends MinValue {
		public MaxValue(HashSet h, Getter g) {super(h,g);}
		public boolean Compare(double x1, double x2) {return (x1 > x2);}
	}
	class EdgeEnum implements Enumeration {
		private Edge next;
		private boolean more = true, last = false;
		private Enumeration enum;
		
		private void getnext() {
			Edge e = next;
			if (enum.hasMoreElements()) {
				next = (Edge)enum.nextElement();		
				while(!inGraph(next) && enum.hasMoreElements()) 
					next = (Edge)enum.nextElement();
				if (inGraph(next)) return;
				next = e;
			}		
			more = false;
			return;
		}
		public EdgeEnum(Graph g)	{
			enum = g.getEdgeEnum();
			getnext();
			}
		public boolean hasMoreElements() {return !last;}
		public Object nextElement() {
			Edge deliver = next;
			if (!more) {
				last = true;
				return deliver;
			}
			getnext();
			return deliver;
		}
	}
	private Graph g;
	private HashSet h = new HashSet();
	private Getx getx = new Getx();
	private Gety gety = new Gety();
	private MaxValue maxX = new MaxValue(h, getx)
	               , maxY = new MaxValue(h, gety);
	private MinValue minX = new MinValue(h, getx)
	               , minY = new MinValue(h, gety);
	
	public static final Object ADDED = new Object();
	public static final Object DELETED = new Object();
	
	/**
	 *Creates a new empty sub-Graph.
	 *@param	g	<code>Graph</code> of which you want to make a sub-Graph.
	 */
	public SubGraph(Graph g) {
		if (g != null) {		
			this.g = g;
		} else 
		System.out.println("Warning : tried to initialize with a unintialized Graph."
		                 + " Init-call ignored!");
	}
	
	private void checkToInvalidate(Node n) {
/*		if (n == maxX.getNode()) maxX.invalidate();
		if (n == maxY.getNode()) maxY.invalidate();
		if (n == minX.getNode()) minX.invalidate();
		if (n == minY.getNode()) minY.invalidate();*/
		maxX.invalidate();
		maxY.invalidate();
		minX.invalidate();
		minY.invalidate();	
	}
	
	/**
	 *Adds a new node into the sub-Graph. 
	 */
	public void addNode(Node n) {
	if (n != null) {
		if (g.inGraph(n)) {
				h.add(n);
				n.addObserver(this);
				maxX.invalidate();
				maxY.invalidate();
				minX.invalidate();
				minY.invalidate();
				notifyObservers(ADDED);
			} else System.out.println("Warning : tried to add a Node not owned"
			                        + "by the Graph. Method-call ignored!");
		} else System.out.println("Warning : tried to add an unintialized Node."
	                         + " Method-call ignored!");
	}
	
	/**
	 *Deletes a node of the sub-Graph. 
	 */
	public void deleteNode(Node n) {
	if (n != null) {
			h.delete(n);
			checkToInvalidate(n);
			notifyObservers(DELETED);
		} else System.out.println("Warning : tried to delete an unintialized Node."
		                        + " Method-call ignored!");
	}
	
	/**
	 *Checks if the sub-Graphs owns the requested node.
	 */
	public boolean inGraph(Node n) {return h.contains(n);}

	/**
	 *Checks if the sub-Graphs owns both nodes of the requested edge.
	 */
	public boolean inGraph(Edge e) {return (inGraph(e.getn1()) && inGraph(e.getn2()));}
	
	/**
	 *Counts the sub-Graphs nodes.
	 */
	public int getNodeCount()	{return h.getCount();}
	
	/**
	 *Returns the <code>Enumeration</code> of the sub-Graph, containing all
	 *nodes.
	 */
	public Enumeration getNodeEnum()	{return h.getEnum();}
	
	public Enumeration getEdgeEnum()	{return new EdgeEnum(g);}
	
	/**
	 *Traverses all nodes and processes it with the given
	 *<code>NodeTraverser</code>.
	 *@see graphomat.graph.Graph#traversNodes
	 */
	public void traversNodes(NodeTraverser t) {
		Enumeration enum = getNodeEnum();
		while(enum.hasMoreElements()) {
			t.processNode((Node)enum.nextElement());
		}
	}
	
	/**
	 *Is called when a recalculation might be needed.
	 */
	public void update(Observable n, Object arg) {
		if (n instanceof Node) {
			if (arg == Node.DELETED) deleteNode((Node) n);
			if (arg == Node.MOVED) checkToInvalidate((Node) n);
		}
	}
	
	public  void notifyObservers(Object arg) {
		setChanged();
		super.notifyObservers(arg);
	}
	/**
	 *Returns the biggest x-value of the Graph.
	 */
	public double getMaxX()	{return maxX.getValue();}
	
	/**
	 *Returns the biggest y-value of the Graph.
	 */
	public double getMaxY()	{return maxY.getValue();}
	
	/**
	 *Returns the smallest x-value of the Graph.
	 */
	public double getMinX()	{return minX.getValue();}
	
	/**
	 *Returns the smallest y-value of the Graph.
	 */
	public double getMinY()	{return minY.getValue();}
	
	/**
	 *Returns a String containg all information about the sub-Graph.
	 */
	public String toString()	{return h.toString();}
}